
#ifndef  CALIBRATIONALG_H
#define CALIBRATIONALG_H

#include "Headers.h"
#include "Event.h"
#include "Geometry.h"
#include "SystemAlg.h"


/*! -Class:    CalibrationAlg
  - Author:   Tamara Starke and Michel Lefebvre
  - Date:      09/07/2003
  - Description: CalibrationAlg is a simple SystemAlg that allows access to the content of
                 calibration files (calibration coefficients and status).  It also
                 allows access to ped rms in nA
                 Possible future development includes calibrating adc signals
  */
class CalibrationAlg : public SystemAlg {
public:

  /** @param event A pointer in the event. Information contained in the event
    *                             gets refreshed before the execute method is run
    * @param hecGeom Pointer to the singleton Geometry object
    * @param runHeader Pointer to the singleton HecRunHeader object
    */

  CalibrationAlg(Event* event, Geometry* hecGeom, Headers* runHeader);

  //! Gets run once at the begining of the event Loop
  bool initialize();

  //! Gets run for every event
  bool execute();

  //! Gets run after all events have been executed
  bool finalize();

  //===== hec public methods =====

  //! Load hec calibration coefficients from file and fill the ped rms in nA map
  bool hecLoad(string hecCalibFileName);

  //! Get febno keyed hec calibration coefficient map
  map<int, vector<float> > hecCoeffMap();

  //! Get febno keyed hec calibration coefficient status map
  map<int, int> hecCoeffStatusMap();

  //! Get hec calibration coefficients
  vector<float> hecCoeff(int febno);

  //! Get hec calibration coefficient status
  int hecCoeffStatus(int febno);

  //! Get febno keyed hec ped rms in nA map
  map<int, float> hecPedRMSnAMap();

  //! Get hec ped rms in nA
  float hecPedRMSnA(int febno);
  
  //! pedestal subtracted ADC to nA conversion for the hec
  float hecADCTonA(int febno, float ADC);
 

  //===== emec public methods =====

  //! Load emec calibration coefficients from file and fill the ped rms in nA map
  bool emecLoad(string emecCalibFileName);

  //! Get febno keyed emec calibration coefficient map
  map<int, vector<float> > emecCoeffMap();

  //! Get febno keyed emec calibration coefficient status map
  map<int, int> emecCoeffStatusMap();

  //! Get emec calibration coefficients
  vector<float> emecCoeff(int febno);

  //! Get emec calibration coefficient status
  int emecCoeffStatus(int febno);

  //! Get febno keyed emec ped rms in nA map
  map<int, float> emecPedRMSnAMap();

  //! Get emec ped rms in nA
  float emecPedRMSnA(int febno);

  //! pedestal subtracted ADC to nA conversion for the emec
  float emecADCTonA(int febno, float ADC);

private:
  Geometry* m_geo;
  Event* m_event;
  Headers* m_runHeader;

  //################################
  //### insert your member variables here

  // hec calibration file version number
  int m_hecVersion;
  // hec number of calibration coefficients per channel
  int m_hecNumberOfCoeff;
  // map of febno keyed vector of hec calibration coefficients
  map<int, vector<float> > m_hecCoeffMap;
  // map of febno keyed hec calibration coefficients status
  map<int, int> m_hecCoeffStatusMap;
  // map of febno keyed hec pedestal rms in nA
  map<int, float> m_hecPedRMSnAMap;


  // emec calibration file version number
  int m_emecVersion;
  // emec number of calibration coefficients per channel
  int m_emecNumberOfCoeff;
  // map of febno keyed vector of emec calibration coefficients
  map<int, vector<float> > m_emecCoeffMap;
  // map of febno keyed emec calibration coefficients status
  map<int, int> m_emecCoeffStatusMap;
  // map of febno keyed emec pedestal rms in nA
  map<int, float> m_emecPedRMSnAMap;


  //################################
};

#endif
