
#ifndef  CELLCORRECTIONALG_H
#define CELLCORRECTIONALG_H

#include "Headers.h"
#include "Event.h"
#include "Geometry.h"
#include "SystemAlg.h"

#include "NoiseAlg.h"
#include "CalibrationAlg.h"

#include <string>
#include <map>

using std::string;
using std::map;

/*! -Class:   CellCorrectionAlg
  - Author:   Michel Lefebvre
  - Date:     13/8/2003
  - Description: CellCorrectionAlg is a systemAlg that computes cell signal correction factors
                 - for HV effects
                 - for sampling fraction effects
                 - for global timing effects
                 Corrected signal is obtained by multiplying the signal by the correction factors
  */
class CellCorrectionAlg : public SystemAlg {
public:

  /** The user should never need to use this contructor
    * @param event A pointer in the event. Information contained in the event
    *                             gets refreshed before the execute method is run
    * @param hecGeom Pointer to the singleton Geometry object
    * @param runHeader Pointer to the singleton HecRunHeader object
    */

  CellCorrectionAlg(Event* event, Geometry* hecGeom, Headers* runHeader);

  //! Gets run once at the begining of the event Loop
  bool initialize();

  //! Gets run for every event
  bool execute();

  //! Gets run after all events have been executed
  bool finalize();

  //################################
  // hec public methods

  /** Returns a correction factor that corrects for known HV problems
    * @param febno channel febno
    */
  float hecHVCorrection(int febno);

  /** Returns a correction factor that corrects for sampling fraction
    * The HEC first wheel is associated with a factor of 1.
    * @param febno channel febno
    */
  float hecSamplingFractionCorrection(int febno);

  /** Returns a correction factor that corrects for global time
    * (TDC or global cubic) signal dependency due to imperfect digital
    * filtering weights.
    * @param febno channel febno
    */
  float hecGlobalTimeCorrection(int febno);

  //################################
  // emec public methods

  /** Returns a correction factor that corrects for known HV problems
    * and/or emec HV features
    * @param febno channel febno
    */
  float emecHVCorrection(int febno);

  /** Returns a correction factor that corrects for sampling fraction
    * @param febno channel febno
    */
  float emecSamplingFractionCorrection(int febno);

  /** Returns a correction factor that corrects for global time
    * (TDC or global cubic) signal dependency due to imperfect digital
    * filtering weights.
    * @param febno channel febno
    */
  float emecGlobalTimeCorrection(int febno);

  /** Returns the LAr gap thickness (cm) in the emec for a given rho (cm)
    * @param rho distance to the z axis of the local detector coordinate system
    */
  float emecGapThickness(float rho);

  //################################
  // other public methods

  /** Computes the electron drift velocity (cm/ns) in LAr given
    * the electric field stregth (kV/cm)
    * @param EField electric field stregth (kV/cm)
    */
  float LArDriftVelocity(float EField);

private:

  //################################
  // private variables

  Geometry* m_geo;
  Event* m_event;
  Headers* m_runHeader;

  //################################
  // other private variables and methods

};

#endif
