#ifndef  CLUSTERINGALG_H
#define CLUSTERINGALG_H

#include "Headers.h"
#include "Event.h"
#include "Geometry.h"
#include "SystemAlg.h"

#include "NoiseAlg.h"
#include "CalibrationAlg.h"

#include <map>
#include <string>
#include <vector>

using std::map;
using std::string;
using std::vector;

/*! -Class:   ClusteringAlg
  - Author:   Michel Lefebvre
  - Date:     28/04/2004
  - Description: ClusteringAlg is a class designed to find HEC and EMEC cell clusters
                 on an event by event basis
                 This is an implementation of Sven Menke's CaloRec/CaloTopoClusterMaker
                 The first seed cell list is in decreasing signal/noise order
                 Optionally, the seed cell list can be ordered for the other iterations

Steps 1 to 3 are done in one loop over all cells
1) if cell above seed threshold (E > 4 sigma) 
   put in list C (cell above seed threshold list)
   and associate it to a cluster
2) else if cell above neighbor threshold (|E| > 3 sigma) 
   put in list B (cell above neighbor threshold list)
3) else if cell above cell threshold (|E| > 2 sigma) 
   put in list A (cell above cell threshold list)
At this point, lists A, B and C are mutually exclusive

4) create an empty list D (seed cell list)
   create list F (next seed cell list) and set it to list C
   sort list F in decreasing order of E/sigma

While there are cells in list F:

5) swap lists D and F, then clear list F
   optionally sort list D in decreasing order of E/sigma

6) loop over the seed cells in list D
   these cells are each associated to a cluster
   
  7) look up the seed cell neighbors
     loop over the seed cell neighbors

    8) if neighbor in list A
         if neighbor not associated to a cluster then add to cluster
    9) else if neighbor in list B
         if neighbor not associated to a cluster 
            then add to cluster and add to list F
         else if neighbor associated to a different cluster
            then merge clusters
   10) else if neighbor in list C
         if neighbor associated to a different cluster then merge clusters

If there are cells in list F, then go back to step 5)

  */
class ClusteringAlg : public SystemAlg {
public:

  /** The user should never need to use this contructor
    * @param event A pointer in the event. Information contained in the event
    *                             gets refreshed before the execute method is run
    * @param hecGeom Pointer to the singleton Geometry object
    * @param runHeader Pointer to the singleton HecRunHeader object
    */

  ClusteringAlg(Event* event, Geometry* hecGeom, Headers* runHeader);

  //! Gets run once at the begining of the event Loop
  bool initialize();

  //! Gets run for every event
  bool execute();

  //! Gets run after all events have been executed
  bool finalize();


  /** hec: enable the search of clusters
    * If no noise file name is given, then the file is assumed to be
    * hec_nnnnn_noise.dat
    * @param noiseFileName Name of noise file 
    * @param clusteringIs2D type of clustering: true = 2D, false = 3D
    * @param sortingAtEachIteration if true, sorting of seed cells (S/N) is done at each iteration
    *                               if false, sorting of seed cells (S/N) is done for first iteration only
    * @param cellThreshold absolute energy cell threshold in units of sigma noise
    * @param neighborThreshold absolute energy neighbor threshold in units of sigma noise
    * @param seedThreshold energy seed threshold in units of sigma noise
    */
  void hecEnable(string noiseFileName = "", bool clusteringIs2D = true, bool sortingAtEachIteration = false, float cellThreshold = 2., float neighborThreshold = 3., float seedThreshold = 4.);

  /** hec: set the calibration file name
    * Needed if the ntuple data is in ADC 
    * Must be called before hecEnable 
    * @param calibrationFileName the name of the calibration file
    */
  void hecCalibrationFileName(string calibrationFileName){m_hecCalibrationFileName = calibrationFileName;};

  //! hec: get number of iterations needed in the clustering algorithm
  int getHecNumberOfIterations() {return m_hecNumberOfIterations;};

  //! hec: get the cluster ID map, febno keyed
  map<int, int> getHecClusterIDMap() {return m_hecClusterIDMap;};

  //! hec: get the cluster map, clusterID keyed.  A Cluster is a vector<int>
  map<int, vector<int> > getHecClusterMap() {return m_hecClusterMap;};

  //! hec: get absolute energy cell threshold in units of sigma noise
  float getHecCellThrehold() {return m_hecCellThreshold;};

  //! hec: get absolute energy neighbor threshold in units of sigma noise
  float getHecNeighborThrehold() {return m_hecNeighborThreshold;};

  //! hec: get energy seed threshold in units of sigma noise
  float getHecSeedThrehold() {return m_hecSeedThreshold;};


  /** emec: enable the search of clusters
    * If no noise file name is given, then the file is assumed to be
    * emec_nnnnn_noise.dat
    * @param noiseFileName Name of noise file 
    * @param clusteringIs2D type of clustering: true = 2D, false = 3D
    * @param sortingAtEachIteration if true, sorting of seed cells (S/N) is done at each iteration
    *                               if false, sorting of seed cells (S/N) is done for first iteration only
    * @param cellThreshold absolute energy cell threshold in units of sigma noise
    * @param neighborThreshold absolute energy neighbor threshold in units of sigma noise
    * @param seedThreshold energy seed threshold in units of sigma noise
    */
  void emecEnable(string noiseFileName = "", bool clusteringIs2D = true, bool sortingAtEachIteration = false, float cellThreshold = 2., float neighborThreshold = 3., float seedThreshold = 4.);

  /** emec: set the calibration file name
    * Needed if the ntuple data is in ADC 
    * Must be called before emecEnable 
    * @param calibrationFileName the name of the calibration file
    */
  void emecCalibrationFileName(string calibrationFileName){m_emecCalibrationFileName = calibrationFileName;};

  //! emec: get number of iterations needed in the clustering algorithm
  int getEmecNumberOfIterations() {return m_emecNumberOfIterations;};

  //! emec: get the cluster ID map, febno keyed
  map<int, int> getEmecClusterIDMap() {return m_emecClusterIDMap;};

  //! emec: get the cluster map, clusterID keyed.  A Cluster is a vector<int>
  map<int, vector<int> > getEmecClusterMap() {return m_emecClusterMap;};

  //! emec: get absolute energy cell threshold in units of sigma noise
  float getEmecCellThrehold() {return m_emecCellThreshold;};

  //! emec: get absolute energy neighbor threshold in units of sigma noise
  float getEmecNeighborThrehold() {return m_emecNeighborThreshold;};

  //! emec: get energy seed threshold in units of sigma noise
  float getEmecSeedThrehold() {return m_emecSeedThreshold;};


private:
  //################################
  // private methods

  //! clears the vector of clusters and the cluster ID map
  void clearClusters();

  //! create a cluster, and return its ID
  int createCluster();

  //! add a cell to a cluster
  void addToCluster(int clusterID, int febno);

  //! merge cluster2 into cluster1, delete cluster2
  void mergeClusters(int clusterID1, int clusterID2);

  //! set a febno as not belonging to any cluster
  void isNotInAnyCluster(int febno) {m_clusterIDMap[febno] = -1;};

  //! true if a febno is not belonging to any cluster
  bool notInAnyCluster(int febno) {return m_clusterIDMap[febno] < 0;};

  //################################
  // private variables

  Geometry* m_geo;
  Event* m_event;
  Headers* m_runHeader;
  // pointer to NoiseAlg systemAlg
  NoiseAlg* m_noise;
  // pointer to CalibrationAlg systemAlg
  CalibrationAlg* m_calibration;

  // first event
  bool m_firstEvent;
  // next available cluster ID
  int m_clusterID;
  // cluster ID map, febno keyed
  map<int, int> m_clusterIDMap;
  // cluster map, cluster ID keyed
  // where a cluster is a vector of febnos
  map<int, vector<int> > m_clusterMap;

  // hec
  bool m_hecEnable;
  bool m_hecClusteringIs2D;
  float m_hecCellThreshold;
  float m_hecNeighborThreshold;
  float m_hecSeedThreshold;
  bool m_hecSortingAtEachIteration;
  string m_hecCalibrationFileName;
  int m_hecNumberOfIterations;
  // cluster ID map, febno keyed
  map<int, int> m_hecClusterIDMap;
  // cluster map, cluster ID keyed
  // where a cluster is a vector of febnos
  map<int, vector<int> > m_hecClusterMap;

  // emec
  bool m_emecEnable;
  bool m_emecClusteringIs2D;
  float m_emecCellThreshold;
  float m_emecNeighborThreshold;
  float m_emecSeedThreshold;
  bool m_emecSortingAtEachIteration;
  string m_emecCalibrationFileName;
  int m_emecNumberOfIterations;
  // cluster ID map, febno keyed
  map<int, int> m_emecClusterIDMap;
  // cluster map, cluster ID keyed
  // where a cluster is a vector of febnos
  map<int, vector<int> > m_emecClusterMap;

};

#endif
