
#include "CellCorrectionAlg.h"

#include <iostream>
#include <cmath>
#include <vector>

using std::cout;
using std::endl;
using std::vector;

CellCorrectionAlg::CellCorrectionAlg(Event* event, Geometry* geo, Headers* runHeader):SystemAlg( event,geo,runHeader){
  m_event = event;
  m_geo = geo;
  m_runHeader = runHeader;

  // create the new System Alg Objects
  vector<SystemAlg*> algVector;

  SystemAlg::setAlgs(algVector);
}


// initialize any variables here
bool CellCorrectionAlg::initialize(){
  SystemAlg::initialize();

  return true;
}

// put any code you want executed per event here
bool CellCorrectionAlg::execute(){
  SystemAlg::execute();

  return true;
}

//put any code here that you want executed after all the events have been processed.
bool CellCorrectionAlg::finalize(){
  SystemAlg::finalize();//do NOT remove this line

  return true;
}

//################################
// hec public methods

// Returns a correction factor that corrects for known HV problems
float CellCorrectionAlg::hecHVCorrection(int febno) {
  if (m_geo->hec_isConnected(febno)) {
    float factor = 1.;
    // use run number 
    // int runNumber =  m_runHeader->global()->getRunNo();
    return factor;
  } else {
    cout << "febno not connected in hecHVCorrection(): " << febno << endl;
    exit(1);
  }
}

// Returns a correction factor that corrects for sampling fraction
// A factor of 1 is associated to the HEC first wheel
// A nominal factor of two is associated to the second wheel
float CellCorrectionAlg::hecSamplingFractionCorrection(int febno) {
  if (m_geo->hec_isConnected(febno)) {
    if (m_geo->hec_iz(febno) < 3) {  // first wheel
      return 1.;
    } else { // second wheel
      return 2.;
    }
  } else {
    cout << "febno not connected in hecSamplingFractionCorrection(): " << febno << endl;
    exit(1);
  }
}
   
// Returns a correction factor that corrects for global time
// (TDC or global cubic) signal dependency due to imperfect digital
// filtering weights.
float CellCorrectionAlg::hecGlobalTimeCorrection(int febno) {
  if (m_geo->hec_isConnected(febno)) {
    float factor = 1.;
    if (m_runHeader->hec()->getTimingType() > 0) { // global cubic time
      // beware of bad cubic global timing
      // if (m_event->globalCubicTimeSigma() == 0) { // bad global cubic timing
      //
    } else { // TDC global time
      //
    }
    //
    return factor;
  } else {
    cout << "febno not connected in hecGlobalTimeCorrection(): " << febno << endl;
    exit(1);
  }
}

//################################
// emec public methods

// Returns a correction factor that corrects for known HV problems
// and/or emec HV features
float CellCorrectionAlg::emecHVCorrection(int febno) {
  if (m_geo->emec_isConnected(febno)) {
    float factor = 1.;
    // assume the relevant rho is that of the median center of the channel
    float rho = m_geo->emec_LocalCenter(febno).Perp();
    float eta = m_geo->emec_eta(febno); // or m_geo->emec_LocalCenter(febno).Eta();
    // choose a rho for which factor = 1 
    // Since the signal depends on the linear charge density times the drift velocity,
    // it would be good to ensure that factor = 1 for the rho's where EField = 10kV/cm, 
    // or any other convenient EField
    // To compute the factor you will need to use
    // emecGapThickness(float rho) and the knowledge of Voltage(eta)
    // LArDriftVelocity(float EFied) where EField = Voltage/GapThickness
    return factor;
  } else {
    cout << "febno not connected in emecHVCorrection(): " << febno << endl;
    exit(1);
  }
}

// Returns a correction factor that corrects for sampling fraction
float CellCorrectionAlg::emecSamplingFractionCorrection(int febno) {
  if (m_geo->emec_isConnected(febno)) {
    float factor = 1.;
    // assume the relevant rho is that of the median center of the channel
    float rho = m_geo->emec_LocalCenter(febno).Perp();
    float eta = m_geo->emec_eta(febno); // or m_geo->emec_LocalCenter(febno).Eta();
    // choose a rho for which factor = 1 : I propose you use the same
    // as for emec HVCorrection
    return factor;
  } else {
    cout << "febno not connected in emecSamplingFractionCorrection(): " << febno << endl;
    exit(1);
  }
}
   
// Returns a correction factor that corrects for global time
// (TDC or global cubic) signal dependency due to imperfect digital
// filtering weights.
float CellCorrectionAlg::emecGlobalTimeCorrection(int febno) {
  if (m_geo->emec_isConnected(febno)) {
    float factor = 1.;
    if (m_runHeader->emec()->getTimingType() > 0) { // global cubic time
      // beware of bad cubic global timing
      // if (m_event->globalCubicTimeSigma() == 0) { // bad global cubic timing
      //
    } else { // TDC global time
      //
    }
    //
    return factor;
  } else {
    cout << "febno not connected in emecGlobalTimeCorrection(): " << febno << endl;
    exit(1);
  }
}
   
// Returns the LAr gap thickness (cm) in the emec for a given rho (cm)
// where rho is the distance to the z axis of the local detector coordinate system
float CellCorrectionAlg::emecGapThickness(float rho) {
  // parametrize emec gap thickness (in cm) as a function of rho
  float gapThickness = 0.2;
  return gapThickness;
}

//################################
// other public methods

// Computes the electron drift velocity (cm/ns) in LAr given
// the electric field stregth (kV/cm)
float CellCorrectionAlg::LArDriftVelocity(float EField) {
  // parametrize the LAr drift velocity (in cm/ns) as a function of the 
  // electric field strength (kV/cm)
  // assume no O2 equivalent impurity in the LAr
  float driftVelocity = 0.00037; // close to what is should be at 10 kV/cm (I think!)
  return driftVelocity;
}

