
#include "NoiseAlg.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <vector>

using std::cout;
using std::endl;
using std::istringstream;
using std::setw;
using std::vector;

NoiseAlg::NoiseAlg(Event* event, Geometry* geo, Headers* runHeader):SystemAlg( event,geo,runHeader){
  m_event = event;
  m_geo = geo;
  m_runHeader = runHeader;

  // create the new System Alg Objects
  m_calibration = new CalibrationAlg(m_event, m_geo, m_runHeader);
  vector<SystemAlg*> algVector;

  algVector.push_back(m_calibration);
  SystemAlg::setAlgs(algVector);

  // set cell signal range (nA) and number of bins for signal histograms
  m_sigMin = -1000;
  m_sigMax = 1000;
  m_nbins = 100;
  // signal cut variables
  m_SigmaMeanCut = 3.;  // HistoMean-FitMean > m_SigmaMeanCut*FitMeanError 
  m_MeanCut = 0.15;     // HistoMean > m_MeanCut*HistoRMS
  // noise computation enable flag
  m_hecEnable = false;
  m_emecEnable = false;
  // print level
  m_hecPrintLevel = 1;
  m_emecPrintLevel = 1;
  // noise result file name
  m_hecNoiseFileName = "";
  m_emecNoiseFileName = "";
  // run number read in a noise file
  m_hecReadRun = 0;
  m_emecReadRun = 0;
}

NoiseAlg::~NoiseAlg() {
  // delete ROOT objects
//   for (map<int, TH1F*>::iterator it = m_hecHMap.begin(); it != m_hecHMap.end(); ++it) {
//     delete it->second;
//   }
//   for (map<int, TH1F*>::iterator it = m_emecHMap.begin(); it != m_emecHMap.end(); ++it) {
//     delete it->second;
//   }
//   for (vector<TF1*>::iterator it = m_TF1Vector.begin(); it != m_TF1Vector.end(); ++it) {
//     delete (*it);
//   }
}

// initialize any variables here
bool NoiseAlg::initialize(){
  SystemAlg::initialize(); // could be remove since CalibrationAlg does nothing here

  return true;
}

// put any code you want executed per event here
bool NoiseAlg::execute(){
  SystemAlg::execute(); // could be remove since CalibrationAlg does nothing here

  if (!m_hecEnable && !m_emecEnable) return true;

  // only keep physics trigger events (may not be necessary...)
  // Triggers (physics,electron,muon,pi,random)
  // Particle (fELECTRON=1, fMUON=2, fPION=3)
  if (m_event->getHecTriggerBit(0) == 0 || m_event->getHecTriggerBit(4) == 1) {  // not physics or a random
    return true;
  }

  // in the case of global cubic fit timing, only keep good events
  if (m_runHeader->hec()->getTimingType() > 0) { // global cubic time
    if (m_event->globalCubicTimeSigma() == 0) { // bad global timing
      return true;
    }
  }

  // fill cell signal histograms
  // if data in ADC, convert to nA
  if (m_hecEnable) {
    for (map<int, TH1F*>::iterator it = m_hecHMap.begin(); it != m_hecHMap.end(); ++it) {
      int febno = it->first;
      float s = m_event->hecCell(febno)->signal();
      if (m_runHeader->hec()->getEUnit() == 1) s = m_calibration->hecADCTonA(febno, s);  // ADC to nA
      m_hecHMap[febno]->Fill(s);
    }
  }
  if (m_emecEnable) {
    for (map<int, TH1F*>::iterator it = m_emecHMap.begin(); it != m_emecHMap.end(); ++it) {
      int febno = it->first;
      float s = m_event->emecCell(febno)->signal();
      if (m_runHeader->emec()->getEUnit() == 1) s = m_calibration->emecADCTonA(febno, s);  // ADC to nA
      m_emecHMap[febno]->Fill(s);
    }
  }

  return true;
}

//put any code here that you want executed after all the events have been processed.
bool NoiseAlg::finalize(){
  SystemAlg::finalize();//do NOT remove this line

  if (!m_hecEnable && !m_emecEnable) return true;

  TBID* ID = new TBID();

  // fitting function
  TF1* fGaussAll = new TF1("fGaussAll", "gaus", m_sigMin, m_sigMax);
  m_TF1Vector.push_back(fGaussAll);

  // ================= hec
  if (m_hecEnable) {
    if (m_hecPrintLevel >= 3) {
      cout << "hec fit results" << endl;
      cout << "febno     pedrms(adc, nA)         fit mean(nA)                 fit sigma(nA)         fit prob" << endl;
    }
    for (map<int, TH1F*>::iterator it = m_hecHMap.begin(); it != m_hecHMap.end(); ++it) {
      int febno = it->first;
      // get histogram signal mean and rms
      m_hecHistoMeanMap[febno] = it->second->GetMean();
      m_hecHistoRMSMap[febno]  = it->second->GetRMS();
      m_hecHistoNMap[febno]    = it->second->Integral(1, it->second->GetNbinsX());
      // fit gaussian to all range
      it->second->Fit("fGaussAll", "QR0");
      // extract fit information
      float hecFitMean       = it->second->GetFunction("fGaussAll")->GetParameter(1);
      float hecFitMeanError  = it->second->GetFunction("fGaussAll")->GetParError(1);
      float hecFitSigma      = it->second->GetFunction("fGaussAll")->GetParameter(2);
      float hecFitSigmaError = it->second->GetFunction("fGaussAll")->GetParError(2);
      float hecFitProb       = it->second->GetFunction("fGaussAll")->GetProb();
      if (m_hecPrintLevel >= 4) {
        cout << setw(5)  << febno
             << setw(10) << m_runHeader->hec()->getPedRms(febno)
             << setw(10) << m_calibration->hecPedRMSnA(febno)
             << setw(12) << hecFitMean << " +/- "
             << setw(10) << hecFitMeanError
             << setw(12) << hecFitSigma << " +/- "
             << setw(10) << hecFitSigmaError
             << setw(15) << hecFitProb << endl;
      }
      // fitting function, with range +/- 2 sigma of previous fit
      TF1* fGauss = new TF1("fGauss", "gaus", hecFitMean-2*hecFitSigma, hecFitMean+2*hecFitSigma); 
      m_TF1Vector.push_back(fGauss);
      it->second->Fit("fGauss", "QR0");
      // extract fit information
      hecFitMean       = it->second->GetFunction("fGauss")->GetParameter(1);
      hecFitMeanError  = it->second->GetFunction("fGauss")->GetParError(1);
      hecFitSigma      = it->second->GetFunction("fGauss")->GetParameter(2);
      hecFitSigmaError = it->second->GetFunction("fGauss")->GetParError(2);
      hecFitProb       = it->second->GetFunction("fGauss")->GetProb();
      if (m_hecPrintLevel >= 3) {
        cout << setw(5)  << febno
             << setw(10) << m_runHeader->hec()->getPedRms(febno)
             << setw(10) << m_calibration->hecPedRMSnA(febno)
             << setw(12) << hecFitMean << " +/- "
             << setw(10) << hecFitMeanError
             << setw(12) << hecFitSigma << " +/- "
             << setw(10) << hecFitSigmaError
             << setw(15) << hecFitProb << endl;
      }
      // extract potentially bad channels for the purpose of noise measurement
      // 1- HistoMean-FitMean > m_SigmaMeanCut*FitMeanError  (signal in channel)
      // 2- HistoMean > m_MeanCut*HistoRMS                   (signal in channel)
      // 3- FitProb <= 0 || FitSigmaError/PedRMS > 1         (very bad fit)
      // 4- HistoRMS == 0                                    (identical entries)
      // 5- HistoN == 0                                      (no entries in histo window)
      int st = 0;
      if (m_hecHistoMeanMap[febno]-hecFitMean > m_SigmaMeanCut*hecFitMeanError) st += 1;
      if (m_hecHistoMeanMap[febno] > m_MeanCut*m_hecHistoRMSMap[febno]) st +=10;
      if (hecFitProb <= 0 || 
          (m_calibration->hecPedRMSnA(febno) > 0 && hecFitSigmaError/m_calibration->hecPedRMSnA(febno) > 1)) st += 100;
      if (m_hecHistoRMSMap[febno] == 0) st += 1000;
      if (m_hecHistoNMap[febno] == 0) st += 10000;
      m_hecNoiseStatusMap[febno] = st;
      if (st == 0) {
        // only keep map entries for good channels
        m_hecFitMeanMap[febno]       = hecFitMean;       
        m_hecFitMeanErrorMap[febno]  = hecFitMeanError;  
        m_hecFitSigmaMap[febno]      = hecFitSigma;      
        m_hecFitSigmaErrorMap[febno] = hecFitSigmaError; 
        m_hecFitProbMap[febno]       = hecFitProb;
        if (m_calibration->hecPedRMSnA(febno) > 0) {
          m_hecFitSigmaOverPedRMSMap[febno] = m_hecFitSigmaMap[febno]/m_calibration->hecPedRMSnA(febno);
          m_hecFitSigmaOverPedRMSErrorMap[febno] = m_hecFitSigmaErrorMap[febno]/m_calibration->hecPedRMSnA(febno);
        }
      }
    }
    // compute the mean and rms of the m_hecFitSigmaOverPedRMS for various iz and ieta
    map<int, int>   hec_m0;  // counter
    map<int, float> hec_m1;  // sum of m_hecFitSigmaOverPedRMS
    map<int, float> hec_m2;  // sum of (m_hecFitSigmaOverPedRMSError)^2
    for (map<int, float>::iterator it = m_hecFitSigmaOverPedRMSMap.begin(); it != m_hecFitSigmaOverPedRMSMap.end(); ++it) {
      int febno = it->first;
      int key = hecMakeTBID(febno).id();
      if (hec_m0.count(key) == 0) {
        hec_m0[key] = 1;
        hec_m1[key] = m_hecFitSigmaOverPedRMSMap[febno];
        hec_m2[key] = pow(m_hecFitSigmaOverPedRMSErrorMap[febno],2);  // error on mean
      } else {
        hec_m0[key] += 1;
        hec_m1[key] += m_hecFitSigmaOverPedRMSMap[febno];
        hec_m2[key] += pow(m_hecFitSigmaOverPedRMSErrorMap[febno],2);
      }
    }
    map<int, float> hecMeanRatioTempSMap;  // map with special key
    map<int, float> hecMeanRatioErrorTempSMap;
    for (map<int, int>::iterator it = hec_m0.begin(); it != hec_m0.end(); ++it) {
      int key = it->first;
      hecMeanRatioTempSMap[key] = hec_m1[key]/hec_m0[key];
      float var = hec_m2[key]/hec_m0[key];
      hecMeanRatioErrorTempSMap[key] = (var > 0.) ? sqrt(var) : 0.;
    }
    map<int, int> hecKeys;  // all the possible hec keys
    for (int febno = m_runHeader->hec()->getFirstCh(); febno <= m_runHeader->hec()->getLastCh(); ++febno) {
      if (m_geo->hec_isConnected(febno)) {
        int key = hecMakeTBID(febno).id();
        hecKeys[key] = 1;
      }
    }
    // deal with ieta values with no ratio results
    // fill the final mean ratio special map
    map<int, float> hecMeanRatioSMap;  // map with special key
    map<int, float> hecMeanRatioErrorSMap;
    for (map<int, int>::iterator it = hecKeys.begin(); it != hecKeys.end(); ++it) {
      int key = it->first;
      if (hecMeanRatioTempSMap.count(key) == 0) {
        // need to do something about keys that have no mean ratio
        // this is the case if all cells in one eta row are bad cells
        // look for the nearest ieta value with a mean ratio
        int j = 1;
        while(hecMeanRatioSMap.count(key) == 0 && j < 10) {  // try up to the 9th neighbour in ieta
          int key_over = TBID(ID->idet(key), ID->ieta(key)+j, ID->iphi(key), ID->iz(key)).id();
          if (hecMeanRatioTempSMap.count(key_over) > 0) {
            hecMeanRatioSMap[key]      = hecMeanRatioTempSMap[key_over];
            hecMeanRatioErrorSMap[key] = hecMeanRatioErrorTempSMap[key_over];
            hec_m0[key] = 0;
          } else {
            if (ID->ieta(key)-j >= 0) {  // ieta = 0 is safe for all detectors here
              int key_below = TBID(ID->idet(key), ID->ieta(key)-j, ID->iphi(key), ID->iz(key)).id();
              if (hecMeanRatioTempSMap.count(key_below) > 0) {
                hecMeanRatioSMap[key]      = hecMeanRatioTempSMap[key_below];
                hecMeanRatioErrorSMap[key] = hecMeanRatioErrorTempSMap[key_below];
                hec_m0[key] = 0;
              }
            }
          }
          ++j;
        }
      } else {
        // this key has a mean ratio
        hecMeanRatioSMap[key]      = hecMeanRatioTempSMap[key];
        hecMeanRatioErrorSMap[key] = hecMeanRatioErrorTempSMap[key];
      }
    }
    if (m_hecPrintLevel >= 2) {
      cout << "hec sigma/pedrms mean ratios" << endl;
      cout << "number of good channels " << m_hecFitMeanMap.size() << endl;
      cout << "number of bad channels  " << m_hecHMap.size() - m_hecFitMeanMap.size() << endl;
      cout << "ieta  iz           ratio mean           N" << endl;
      for (map<int, int>::iterator it = hecKeys.begin(); it != hecKeys.end(); ++it) {
        int key = it->first;
        if (hecMeanRatioSMap.count(key) == 0) {
          cout << setw(4) << ID->ieta(key) 
               << setw(4) << ID->iz(key) 
               << "no ratios available!" << endl;
        } else {
          cout << setw(4) << ID->ieta(key) 
               << setw(4) << ID->iz(key) 
               << setw(12) << hecMeanRatioSMap[key] << " +/- "
               << setw(12) << hecMeanRatioErrorSMap[key]
               << setw(4) << hec_m0[key] << endl;
        }
      }
    }
    // fill mean ratio maps
    for (map<int, int>::iterator it = m_hecNoiseStatusMap.begin(); it != m_hecNoiseStatusMap.end(); ++it) {
      int febno = it->first;
      int key = hecMakeTBID(febno).id();
      if (hecMeanRatioSMap.count(key) == 0) {
        m_hecMeanRatioMap[febno] = 0.;
        m_hecMeanRatioErrorMap[febno] = 0.;
      } else {
        m_hecMeanRatioMap[febno] = hecMeanRatioSMap[key];
        m_hecMeanRatioErrorMap[febno] = hecMeanRatioErrorSMap[key];
      }
    }
    // fill noise map
    for (map<int, int>::iterator it = m_hecNoiseStatusMap.begin(); it != m_hecNoiseStatusMap.end(); ++it) {
      int febno = it->first;
      if (m_hecNoiseStatusMap[febno] == 0) {
        m_hecNoiseMap[febno] = m_hecFitSigmaMap[febno];
        m_hecNoiseErrorMap[febno] = m_hecFitSigmaErrorMap[febno];
      } else {
        int key = hecMakeTBID(febno).id();
        if (hecMeanRatioSMap.count(key) == 0) {
          cout << "Warning: could not estimate the noise for hec channel febno " << febno << endl;
        }
        m_hecNoiseMap[febno] = m_hecMeanRatioMap[febno] * m_calibration->hecPedRMSnA(febno);
        m_hecNoiseErrorMap[febno] = m_hecMeanRatioErrorMap[febno] * m_calibration->hecPedRMSnA(febno);
      }
    }
    if (m_hecPrintLevel >= 1) hecPrintNoise();
    // write noise result to file
    hecWriteNoiseFile();
  }

  // ================= emec
  if (m_emecEnable) {
    if (m_emecPrintLevel >= 3) {
      cout << "emec fit results" << endl;
      cout << "febno     pedrms(adc, nA)         fit mean(nA)                 fit sigma(nA)         fit prob" << endl;
    }
    for (map<int, TH1F*>::iterator it = m_emecHMap.begin(); it != m_emecHMap.end(); ++it) {
      int febno = it->first;
      // get histogram signal mean and rms
      m_emecHistoMeanMap[febno] = it->second->GetMean();
      m_emecHistoRMSMap[febno]  = it->second->GetRMS();
      m_emecHistoNMap[febno]    = it->second->GetSumOfWeights();
      // fit gaussian to all range
      it->second->Fit("fGaussAll", "QR0");
      // extract fit information
      float emecFitMean       = it->second->GetFunction("fGaussAll")->GetParameter(1);
      float emecFitMeanError  = it->second->GetFunction("fGaussAll")->GetParError(1);
      float emecFitSigma      = it->second->GetFunction("fGaussAll")->GetParameter(2);
      float emecFitSigmaError = it->second->GetFunction("fGaussAll")->GetParError(2);
      float emecFitProb       = it->second->GetFunction("fGaussAll")->GetProb();
      if (m_emecPrintLevel >= 4) {
        cout << setw(5)  << febno
             << setw(10) << m_runHeader->emec()->getPedRms(febno)
             << setw(10) << m_calibration->emecPedRMSnA(febno)
             << setw(12) << emecFitMean << " +/- "
             << setw(10) << emecFitMeanError
             << setw(12) << emecFitSigma << " +/- "
             << setw(10) << emecFitSigmaError
             << setw(15) << emecFitProb << endl;
      }
      // fitting function, with range +/- 2 sigma of previous fit
      TF1* fGauss = new TF1("fGauss", "gaus", emecFitMean-2*emecFitSigma, emecFitMean+2*emecFitSigma); 
      m_TF1Vector.push_back(fGauss);
      it->second->Fit("fGauss", "QR0");
      // extract fit information
      emecFitMean       = it->second->GetFunction("fGauss")->GetParameter(1);
      emecFitMeanError  = it->second->GetFunction("fGauss")->GetParError(1);
      emecFitSigma      = it->second->GetFunction("fGauss")->GetParameter(2);
      emecFitSigmaError = it->second->GetFunction("fGauss")->GetParError(2);
      emecFitProb       = it->second->GetFunction("fGauss")->GetProb();
      if (m_emecPrintLevel >= 3) {
        cout << setw(5)  << febno
             << setw(10) << m_runHeader->emec()->getPedRms(febno)
             << setw(10) << m_calibration->emecPedRMSnA(febno)
             << setw(12) << emecFitMean << " +/- "
             << setw(10) << emecFitMeanError
             << setw(12) << emecFitSigma << " +/- "
             << setw(10) << emecFitSigmaError
             << setw(15) << emecFitProb << endl;
      }
      // extract potentially bad channels for the purpose of noise measurement
      // 1- HistoMean-FitMean > m_SigmaMeanCut*FitMeanError  (signal in channel)
      // 2- HistoMean > m_MeanCut*HistoRMS                   (signal in channel)
      // 3- FitProb <= 0 || FitSigmaError/PedRMS > 1         (very bad fit)
      // 4- HistoRMS == 0                                    (identical entries)
      // 5- HistoN == 0                                      (no entries in histo window)
      int st = 0;
      if (m_emecHistoMeanMap[febno]-emecFitMean > m_SigmaMeanCut*emecFitMeanError) st += 1;
      if (m_emecHistoMeanMap[febno] > m_MeanCut*m_emecHistoRMSMap[febno]) st +=10;
      if (emecFitProb <= 0 || 
          (m_calibration->emecPedRMSnA(febno) > 0 && emecFitSigmaError/m_calibration->emecPedRMSnA(febno) > 1)) st += 100;
      if (m_emecHistoRMSMap[febno] == 0) st += 1000;
      if (m_emecHistoNMap[febno] == 0) st += 10000;
      m_emecNoiseStatusMap[febno] = st;
      if (st == 0) {
        // only keep map entries for good channels
        m_emecFitMeanMap[febno]       = emecFitMean;       
        m_emecFitMeanErrorMap[febno]  = emecFitMeanError;  
        m_emecFitSigmaMap[febno]      = emecFitSigma;      
        m_emecFitSigmaErrorMap[febno] = emecFitSigmaError; 
        m_emecFitProbMap[febno]       = emecFitProb;
        if (m_calibration->emecPedRMSnA(febno) > 0) {
          m_emecFitSigmaOverPedRMSMap[febno] = m_emecFitSigmaMap[febno]/m_calibration->emecPedRMSnA(febno);
          m_emecFitSigmaOverPedRMSErrorMap[febno] = m_emecFitSigmaErrorMap[febno]/m_calibration->emecPedRMSnA(febno);
        }
      }
    }
    // compute the mean and rms of the m_emecFitSigmaOverPedRMS for various iz and ieta
    map<int, int>   emec_m0;  // counter
    map<int, float> emec_m1;  // sum of m_emecFitSigmaOverPedRMS
    map<int, float> emec_m2;  // sum of (m_emecFitSigmaOverPedRMSError)^2
    for (map<int, float>::iterator it = m_emecFitSigmaOverPedRMSMap.begin(); it != m_emecFitSigmaOverPedRMSMap.end(); ++it) {
      int febno = it->first;
      int key = emecMakeTBID(febno).id();
      if (emec_m0.count(key) == 0) {
        emec_m0[key] = 1;
        emec_m1[key] = m_emecFitSigmaOverPedRMSMap[febno];
        emec_m2[key] = pow(m_emecFitSigmaOverPedRMSErrorMap[febno],2);  // error on mean
      } else {
        emec_m0[key] += 1;
        emec_m1[key] += m_emecFitSigmaOverPedRMSMap[febno];
        emec_m2[key] += pow(m_emecFitSigmaOverPedRMSErrorMap[febno],2);
      }
    }
    map<int, float> emecMeanRatioTempSMap;  // map with special key
    map<int, float> emecMeanRatioErrorTempSMap;
    for (map<int, int>::iterator it = emec_m0.begin(); it != emec_m0.end(); ++it) {
      int key = it->first;
      emecMeanRatioTempSMap[key] = emec_m1[key]/emec_m0[key];
      float var = emec_m2[key]/emec_m0[key];
      emecMeanRatioErrorTempSMap[key] = (var > 0.) ? sqrt(var) : 0.;
    }
    map<int, int> emecKeys;  // all the possible emec keys
    for (int febno = m_runHeader->emec()->getFirstCh(); febno <= m_runHeader->emec()->getLastCh(); ++febno) {
      if (m_geo->emec_isConnected(febno)) {
        int key = emecMakeTBID(febno).id();
        emecKeys[key] = 1;
      }
    }
    // deal with ieta values with no ratio results
    // fill the final mean ratio special map
    map<int, float> emecMeanRatioSMap;  // map with special key
    map<int, float> emecMeanRatioErrorSMap;
    for (map<int, int>::iterator it = emecKeys.begin(); it != emecKeys.end(); ++it) {
      int key = it->first;
      if (emecMeanRatioTempSMap.count(key) == 0) {
        // need to do something about keys that have no mean ratio
        // this is the case if all cells in one eta row are bad cells
        // look for the nearest ieta value with a mean ratio
        int j = 1;
        while(emecMeanRatioSMap.count(key) == 0 && j < 10) {  // try up to the 9th neighbour in ieta
          int key_over = TBID(ID->idet(key), ID->ieta(key)+j, ID->iphi(key), ID->iz(key)).id();
          if (emecMeanRatioTempSMap.count(key_over) > 0) {
            emecMeanRatioSMap[key]      = emecMeanRatioTempSMap[key_over];
            emecMeanRatioErrorSMap[key] = emecMeanRatioErrorTempSMap[key_over];
            emec_m0[key] = 0;
          } else {
            if (ID->ieta(key)-j >= 0) {  // ieta = 0 is safe for all detectors here
              int key_below = TBID(ID->idet(key), ID->ieta(key)-j, ID->iphi(key), ID->iz(key)).id();
              if (emecMeanRatioTempSMap.count(key_below) > 0) {
                emecMeanRatioSMap[key]      = emecMeanRatioTempSMap[key_below];
                emecMeanRatioErrorSMap[key] = emecMeanRatioErrorTempSMap[key_below];
                emec_m0[key] = 0;
              }
            }
          }
          ++j;
        }
      } else {
        // this key has a mean ratio
        emecMeanRatioSMap[key]      = emecMeanRatioTempSMap[key];
        emecMeanRatioErrorSMap[key] = emecMeanRatioErrorTempSMap[key];
      }
    }
    if (m_emecPrintLevel >= 2) {
      cout << "emec sigma/pedrms mean ratios" << endl;
      cout << "number of good channels " << m_emecFitMeanMap.size() << endl;
      cout << "number of bad channels  " << m_emecHMap.size() - m_emecFitMeanMap.size() << endl;
      cout << "ieta  iz           ratio mean           N" << endl;
      for (map<int, int>::iterator it = emecKeys.begin(); it != emecKeys.end(); ++it) {
        int key = it->first;
        if (emecMeanRatioSMap.count(key) == 0) {
          cout << setw(4) << ID->ieta(key) 
               << setw(4) << ID->iz(key) 
               << "no ratios available!" << endl;
        } else {
          cout << setw(4) << ID->ieta(key) 
               << setw(4) << ID->iz(key) 
               << setw(12) << emecMeanRatioSMap[key] << " +/- "
               << setw(12) << emecMeanRatioErrorSMap[key]
               << setw(4) << emec_m0[key] << endl;
        }
      }
    }
    // fill mean ratio maps
    for (map<int, int>::iterator it = m_emecNoiseStatusMap.begin(); it != m_emecNoiseStatusMap.end(); ++it) {
      int febno = it->first;
      int key = emecMakeTBID(febno).id();
      if (emecMeanRatioSMap.count(key) == 0) {
        m_emecMeanRatioMap[febno] = 0.;
        m_emecMeanRatioErrorMap[febno] = 0.;
      } else {
        m_emecMeanRatioMap[febno] = emecMeanRatioSMap[key];
        m_emecMeanRatioErrorMap[febno] = emecMeanRatioErrorSMap[key];
      }
    }
    // fill noise map
    for (map<int, int>::iterator it = m_emecNoiseStatusMap.begin(); it != m_emecNoiseStatusMap.end(); ++it) {
      int febno = it->first;
      if (m_emecNoiseStatusMap[febno] == 0) {
        m_emecNoiseMap[febno] = m_emecFitSigmaMap[febno];
        m_emecNoiseErrorMap[febno] = m_emecFitSigmaErrorMap[febno];
      } else {
        int key = emecMakeTBID(febno).id();
        if (emecMeanRatioSMap.count(key) == 0) {
          cout << "Warning: could not estimate the noise for emec channel febno " << febno << endl;
        }
        m_emecNoiseMap[febno] = m_emecMeanRatioMap[febno] * m_calibration->emecPedRMSnA(febno);
        m_emecNoiseErrorMap[febno] = m_emecMeanRatioErrorMap[febno] * m_calibration->emecPedRMSnA(febno);
      }
    }
    if (m_emecPrintLevel >= 1) emecPrintNoise();
    // write noise result to file
    emecWriteNoiseFile();
  }

  // clean up
  delete ID;

  return true;
}

//################################
// hec public methods

// enable the computation of noise for the hec
void NoiseAlg::hecEnable(string calibFileName){
  // set the enable flag
  m_hecEnable = true;
  // load the calibration files and fill the ped rms in nA maps
  m_calibration->hecLoad(calibFileName);
  // book cell signal histograms for connected channels only
  for (int febno = m_runHeader->hec()->getFirstCh(); febno <= m_runHeader->hec()->getLastCh(); ++febno) {
    if (m_geo->hec_isConnected(febno)) {
      std::ostringstream os;
      string h_title = "run number ";
      os  << h_title << m_runHeader->global()->getRunNo() << "  hec channel " << febno;
      h_title = os.str();
      m_hecHMap[febno] = new TH1F(h_title.c_str(), h_title.c_str(), m_nbins, m_sigMin, m_sigMax);
    }
  }
}

// Get the hec noise (nA)
float NoiseAlg::hecNoise(int febno){
  if (m_geo->hec_isConnected(febno)) {
    return m_hecNoiseMap[febno];
  } else {
    cout << "febno not connected in hecNoise(): " << febno << endl;
    exit(1);
  }
}

// Get the hec noise error estimate (nA)
float NoiseAlg::hecNoiseError(int febno){
  if (m_geo->hec_isConnected(febno) && m_hecNoiseErrorMap.count(febno) == 1) {
    return m_hecNoiseErrorMap[febno];
  } else {
    cout << "febno not in m_hecNoiseErrorMap: " << febno << endl;
    exit(1);
  }
}

// Get the hec noise error estimate (nA)
int NoiseAlg::hecNoiseStatus(int febno){
  if (m_geo->hec_isConnected(febno) && m_hecNoiseStatusMap.count(febno) == 1) {
    return m_hecNoiseStatusMap[febno];
  } else {
    cout << "febno not in m_hecNoiseStatusMap: " << febno << endl;
    exit(1);
  }
}

// Read a hec noise file
void NoiseAlg::hecReadNoiseFile(string noiseFileName){
  // see hecWriteNoiseFile() for the file format
  if (noiseFileName == "") {
    // get the default file name
    // hec_nnnnn_noise.dat
    std::ostringstream os;
    os << "hec_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    noiseFileName = os.str();
  }
  // open the file
  ifstream fStream(noiseFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "hecReadNoiseFile failed to open file:" << noiseFileName << endl;
    exit(1);
  }
  //
  string line;
  istringstream inStream; 
  if (getline(fStream, line)) {  // read line into line
    inStream.clear();
    inStream.str(line);  // use line as stream input
    if (!(inStream >> m_hecReadRun)) {  // read in run number
       cout << "bad format in noise file " << noiseFileName << endl;
       exit(1);
    }
  } else {
    cout << "bad format in noise file " << noiseFileName << endl;
    exit(1);
  }    
  int idet;
  if (getline(fStream, line)) {  // read line into line
    inStream.clear();
    inStream.str(line);  // use line as stream input
    if (!(inStream >> idet)) {  // read in detector number
       cout << "bad format in noise file " << noiseFileName << endl;
       exit(1);
    }
  } else {
    cout << "bad format in noise file " << noiseFileName << endl;
    exit(1);
  }    
  if (idet != 2) { // idet = 2 for the hec
    cout << "Noise file does not contain hec data " << noiseFileName << endl;
    exit(1);
  }
  // clear the noise info maps in case they have already been produced or read
  m_hecNoiseMap.clear();
  m_hecNoiseErrorMap.clear();
  m_hecNoiseStatusMap.clear();
  // read the noise values
  while (getline(fStream, line)) {  // read line into line
    inStream.clear();
    inStream.str(line);  // use line as stream input
    int febno;
    float noise;
    if (inStream >> febno >> noise) {  // read in febno and noise
      if (m_geo->hec_isConnected(febno)) {
        m_hecNoiseMap[febno] = noise;                 // nA
        float noiseError;
        if (inStream >> noiseError) {  // read in noiseError
          m_hecNoiseErrorMap[febno] = noiseError;     // nA
          int noiseStatus;
          if (inStream >> noiseStatus) { // read in noiseStatus
            m_hecNoiseStatusMap[febno] = noiseStatus;
          }
        }
      }
    } else {
      cout << "bad format in noise file " << noiseFileName << endl;
      exit(1);
    }
  }
  fStream.close();
  cout << "hec noise values from run " << m_hecReadRun
       << " read from file " << noiseFileName 
       << " for " << m_hecNoiseMap.size() << " connected channels" << endl;
}

// Print hec noise data
void NoiseAlg::hecPrintNoise(){
  cout << "hec noise" << endl;
  if (m_hecNoiseStatusMap.size() > 0) {
    cout << "   febno     noise(nA)                status code" << endl;
  } else {
    cout << "   febno     noise(nA)" << endl;
  }
  for (map<int, float>::iterator it = m_hecNoiseMap.begin(); it != m_hecNoiseMap.end(); ++it) {
    int febno = it->first;
    cout << setw(8)  << febno
         << setw(12) << m_hecNoiseMap[febno];
    if (m_hecNoiseErrorMap.count(febno) == 1) {
      cout << " +/-" << setw(12) << m_hecNoiseErrorMap[febno];
      if (m_hecNoiseStatusMap.count(febno) == 1) {
        cout << setw(8)  << m_hecNoiseStatusMap[febno];
      }
    }
    cout << endl;
  }
}

//################################
// emec public methods

// enable the computation of noise for the emec
void NoiseAlg::emecEnable(string calibFileName){
  // set the enable flag
  m_emecEnable = true;
  // load the calibration files and fill the ped rms in nA maps
  m_calibration->emecLoad(calibFileName);
  // book cell signal histograms for connected channels only
  for (int febno = m_runHeader->emec()->getFirstCh(); febno <= m_runHeader->emec()->getLastCh(); ++febno) {
    if (m_geo->emec_isConnected(febno)) {
      std::ostringstream os;
      string h_title = "run number ";
      os  << h_title << m_runHeader->global()->getRunNo() << "  emec channel " << febno;
      h_title = os.str();
      m_emecHMap[febno] = new TH1F(h_title.c_str(), h_title.c_str(), m_nbins, m_sigMin, m_sigMax);
    }
  }
}

// Get the emec noise (nA)
float NoiseAlg::emecNoise(int febno){
  if (m_geo->emec_isConnected(febno)) {
    return m_emecNoiseMap[febno];
  } else {
    cout << "febno not connected in emecNoise(): " << febno << endl;
    exit(1);
  }
}

// Get the emec noise error estimate (nA)
float NoiseAlg::emecNoiseError(int febno){
  if (m_geo->emec_isConnected(febno) && m_emecNoiseErrorMap.count(febno) == 1) {
    return m_emecNoiseErrorMap[febno];
  } else {
    cout << "febno not in m_emecNoiseErrorMap: " << febno << endl;
    exit(1);
  }
}

// Get the emec noise error estimate (nA)
int NoiseAlg::emecNoiseStatus(int febno){
  if (m_geo->emec_isConnected(febno) && m_emecNoiseStatusMap.count(febno) == 1) {
    return m_emecNoiseStatusMap[febno];
  } else {
    cout << "febno not in m_emecNoiseStatusMap: " << febno << endl;
    exit(1);
  }
}

// Read a emec noise file
void NoiseAlg::emecReadNoiseFile(string noiseFileName){
  // see emecWriteNoiseFile() for the file format
  if (noiseFileName == "") {
    // get the default file name
    // emech_nnnnn_noise.dat or emecm_nnnnn_noise.dat or emec_nnnnn_noise.dat
    // since the gain is not recorded in the ntuple (!) use the data file name
    string s;
    s.assign(m_runHeader->emec()->getDataFileName(), 0, 5);  // get first 5 characters
    if (s != "emech" && s != "emecm") s = "emec";
    std::ostringstream os;
    os << s+"_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    noiseFileName = os.str();
  }
  // open the file
  ifstream fStream(noiseFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "emecReadNoiseFile failed to open file:" << noiseFileName << endl;
    exit(1);
  }
  //
  string line;
  istringstream inStream; 
  if (getline(fStream, line)) {  // read line into line
    inStream.clear();
    inStream.str(line);  // use line as stream input
    if (!(inStream >> m_emecReadRun)) {  // read in run number
       cout << "bad format in noise file " << noiseFileName << endl;
       exit(1);
    }
  } else {
    cout << "bad format in noise file " << noiseFileName << endl;
    exit(1);
  }    
  int idet;
  if (getline(fStream, line)) {  // read line into line
    inStream.clear();
    inStream.str(line);  // use line as stream input
    if (!(inStream >> idet)) {  // read in detector number
       cout << "bad format in noise file " << noiseFileName << endl;
       exit(1);
    }
  } else {
    cout << "bad format in noise file " << noiseFileName << endl;
    exit(1);
  }    
  if (idet != 1) { // idet = 1 for the emec
    cout << "Noise file does not contain emec data " << noiseFileName << endl;
    exit(1);
  }
  // clear the noise info maps in case they have already been produced or read
  m_emecNoiseMap.clear();
  m_emecNoiseErrorMap.clear();
  m_emecNoiseStatusMap.clear();
  // read the noise values
  while (getline(fStream, line)) {  // read line into line
    inStream.clear();
    inStream.str(line);  // use line as stream input
    int febno;
    float noise;
    if (inStream >> febno >> noise) {  // read in febno and noise
      if (m_geo->emec_isConnected(febno)) {
        m_emecNoiseMap[febno] = noise;                 // nA
        float noiseError;
        if (inStream >> noiseError) {  // read in noiseError
          m_emecNoiseErrorMap[febno] = noiseError;     // nA
          int noiseStatus;
          if (inStream >> noiseStatus) { // read in noiseStatus
            m_emecNoiseStatusMap[febno] = noiseStatus;
          }
        }
      }
    } else {
      cout << "bad format in noise file " << noiseFileName << endl;
      exit(1);
    }
  }
  fStream.close();
  cout << "emec noise values from run " << m_emecReadRun
       << " read from file " << noiseFileName 
       << " for " << m_emecNoiseMap.size() << " connected channels" << endl;
}

// Print emec noise data
void NoiseAlg::emecPrintNoise(){
  cout << "emec noise" << endl;
  if (m_emecNoiseStatusMap.size() > 0) {
    cout << "   febno     noise(nA)                status code" << endl;
  } else {
    cout << "   febno     noise(nA)" << endl;
  }
  for (map<int, float>::iterator it = m_emecNoiseMap.begin(); it != m_emecNoiseMap.end(); ++it) {
    int febno = it->first;
    cout << setw(8)  << febno
         << setw(12) << m_emecNoiseMap[febno];
    if (m_emecNoiseErrorMap.count(febno) == 1) {
      cout << " +/-" << setw(12) << m_emecNoiseErrorMap[febno];
      if (m_emecNoiseStatusMap.count(febno) == 1) {
        cout << setw(8)  << m_emecNoiseStatusMap[febno];
      }
    }
    cout << endl;
  }
}

//################################
// hec private methods

// make a TBID from a febno only considering ieta and iz
// for the hec, use iphi = 1 for iz = 1 or 2
//                  iphi = 2 for iz = 3
TBID NoiseAlg::hecMakeTBID(int febno) {
  int ieta = m_geo->hec_ieta(febno);
  int iz = m_geo->hec_iz(febno);
  int iphi = (iz < 3) ? 1 : 2;
  return TBID(2, ieta, iphi, iz);
}

// write noise data file
void NoiseAlg::hecWriteNoiseFile(){
  // format is
  // line 1:      run number
  // line 2:      idet
  // other lines: febno noise(nA) noiseError(nA) statusCode
  //
  // status code
  // == 0  noise obtained from a +/- 2 sigma fit on the noise
  // != 0  fit not possible for this channel -> noise obtained from the pedestal rms and
  //       the noise/pedrms ratios obtained from other channels
  if (m_hecNoiseFileName == "") {
    // get the default file name
    // hec_nnnnn_noise.dat
    std::ostringstream os;
    os << "hec_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    m_hecNoiseFileName = os.str();
  }
  // open the file
  ofstream fStream(m_hecNoiseFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "hecWriteNoiseFile failed to open file:" << m_hecNoiseFileName << endl;
    exit(1);
  }
  //
  fStream << m_runHeader->global()->getRunNo() << endl;
  fStream << 2 << endl;  // idet = 2 for the hec
  for (map<int, float>::iterator it = m_hecNoiseMap.begin(); it != m_hecNoiseMap.end(); ++it) {
    int febno = it->first;
    fStream << setw(8)  << febno
            << setw(12) << m_hecNoiseMap[febno]
            << setw(12) << m_hecNoiseErrorMap[febno]
            << setw(8)  << m_hecNoiseStatusMap[febno] << endl;
  }
  fStream.close();
  cout << "hec noise values written in file " << m_hecNoiseFileName << endl;
}

//################################
// emec private methods


// make a TBID from a febno only considering ieta and iz
// for the emec, use iphi = 2 for iz = 0 or 1
//                   iphi = 8 for iz = 2
//                   iphi = 0 for iz = 3
TBID NoiseAlg::emecMakeTBID(int febno) {
  int ieta = m_geo->emec_ieta(febno);
  int iz = m_geo->emec_iz(febno);
  int iphi;
  if (iz == 2) {
    iphi = 8;
  } else if (iz == 3) {
    iphi = 0;
  } else {
    iphi = 2;
  }
  return TBID(1, ieta, iphi, iz);
}

// write noise data file
void NoiseAlg::emecWriteNoiseFile(){
  // format is
  // line 1:      run number
  // line 2:      idet
  // other lines: febno noise(nA) noiseError(nA) statusCode
  //
  // status code
  // == 0  noise obtained from a +/- 2 sigma fit on the noise
  // != 0  fit not possible for this channel -> noise obtained from the pedestal rms and
  //       the noise/pedrms ratios obtained from other channels
  if (m_emecNoiseFileName == "") {
    // get the default file name
    // emech_nnnnn_noise.dat or emecm_nnnnn_noise.dat or emec_nnnnn_noise.dat
    // since the gain is not recorded in the ntuple (!) use the data file name
    string s;
    s.assign(m_runHeader->emec()->getDataFileName(), 0, 5);  // get first 5 characters
    if (s != "emech" && s != "emecm") s = "emec";
    std::ostringstream os;
    os << s+"_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    m_emecNoiseFileName = os.str();
  }
  // open the file
  ofstream fStream(m_emecNoiseFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "emecWriteNoiseFile failed to open file:" << m_emecNoiseFileName << endl;
    exit(1);
  }
  //
  fStream << m_runHeader->global()->getRunNo() << endl;
  fStream << 1 << endl;  // idet = 1 for the emec
  for (map<int, float>::iterator it = m_emecNoiseMap.begin(); it != m_emecNoiseMap.end(); ++it) {
    int febno = it->first;
    fStream << setw(8)  << febno
            << setw(12) << m_emecNoiseMap[febno]
            << setw(12) << m_emecNoiseErrorMap[febno]
            << setw(8)  << m_emecNoiseStatusMap[febno] << endl;
  }
  fStream.close();
  cout << "emec noise values written in file " << m_emecNoiseFileName << endl;
}

   
