
#include "RunStatisticsAlg.h"

#include <string>
#include <iomanip>

using std::cout;
using std::setw;
using std::left;
using std::right;
using std::endl;
using std::vector;
using std::string;

RunStatisticsAlg::RunStatisticsAlg(Event* event, Geometry* geo, Headers* runHeader):SystemAlg( event,geo,runHeader){
  m_event = event;
  m_geo = geo;
  m_runHeader = runHeader;

  // create the new System Alg Objects
  vector<SystemAlg*> algVector;

  SystemAlg::setAlgs(algVector);

}


// initialize any variables here
bool RunStatisticsAlg::initialize(){
  SystemAlg::initialize(); // could remove this line

  // clear trigger statistics
  m_triggerStat.clear();
  for (int i = 0; i < 5; ++i) m_triggerStat.push_back(0);
  m_triggers = 0;
  m_badGlobalTiming = 0;
  
  if (m_runHeader->global() == 0) return true;  // no data files

  // timing type
  if (m_runHeader->hec() != 0) { // look at hec timing
    m_timingType = m_runHeader->hec()->getTimingType();
  } else { // look at emec timing instead
    m_timingType = m_runHeader->emec()->getTimingType();
  }
    
  return true;
}

// put any code you want executed per event here
bool RunStatisticsAlg::execute(){
  SystemAlg::execute(); // could remove this line

  // fill trigger statistics
  for (int i = 0; i < 5; ++i) {
    if (m_runHeader->hec() != 0 && m_runHeader->emec() != 0) {
      if (m_event->getHecTriggerBit(i) != m_event->getEmecTriggerBit(i)) {
        cout << "Hec and Emec trigger bits differ!! " << endl;
        exit(1);
      }
    }
    if (m_runHeader->hec() != 0) {
      m_triggerStat[i] += m_event->getHecTriggerBit(i);
    } else {
      m_triggerStat[i] += m_event->getEmecTriggerBit(i);
    }
  }

  ++m_triggers;

  if (m_timingType > 0) { // global cubic time
    if (m_event->globalCubicTimeSigma() == 0) { // bad global timing
      ++m_badGlobalTiming;
    }
  }

  return true;
}

//put any code here that you want executed after all the events have been processed.
bool RunStatisticsAlg::finalize(){
  SystemAlg::finalize(); // could remove this line

  if (m_runHeader->global() == 0) return true;  // no data files

  cout << setw(25) << left << "Run Statistics" << endl;
  cout << setw(25) << left << "run number " 
       << setw(25) << right << m_runHeader->global()->getRunNo() << endl;
  cout << setw(25) << left << "beam energy (GeV) " 
       << setw(25) << right << fabs(m_runHeader->global()->getBeamEnergy()) << endl;
  cout << setw(25) << left << "particle " << setw(25) << right;
  string s = (m_runHeader->global()->getBeamEnergy() > 0) ? "+" : "-";
  if (m_runHeader->global()->getPartType() == 1) {
    cout << "e"+s << endl;
  } else if (m_runHeader->global()->getPartType() == 2) {
    cout << "mu"+s << endl;
  } else if  (m_runHeader->global()->getPartType() == 3) {
    cout << "pi"+s << endl;
  } else {
    cout << "unknown" << endl;
  }

  cout << setw(25) << left << "global timing type " << setw(25) << right;
  if (m_timingType > 0) { 
    cout << "cubic timing" << endl;
  } else {
    cout << "tdc timing" << endl;
  }

  cout << setw(25) << left << "triggers "   
       << setw(25) << right << m_triggers << endl;
  cout << setw(25) << left << "  physics "  
       << setw(25) << right << m_triggerStat[0] << endl;
  cout << setw(25) << left << "  electron " 
       << setw(25) << right << m_triggerStat[1] << endl;
  cout << setw(25) << left << "  muon "     
       << setw(25) << right << m_triggerStat[2] << endl;
  cout << setw(25) << left << "  pion "     
       << setw(25) << right << m_triggerStat[3] << endl;
  cout << setw(25) << left << "  random "   
       << setw(25) << right << m_triggerStat[4] << endl;
  if (m_timingType > 0) {
    cout << setw(25) << left << "  bad global timing "
         << setw(25) << right << m_badGlobalTiming << endl;
  }

  return true;
}

    
